/*----------------------------------------------------------------------------
    S.M.E.L.T. : Small Musically Expressive Laptop Toolkit

    Copyright (c) 2007 Rebecca Fiebrink and Ge Wang.  All rights reserved.
      http://smelt.cs.princeton.edu/
      http://soundlab.cs.princeton.edu/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: smack-sense.c
// desc: Sudden Motion Sensor (SMS)-based smack detector
//
// adapted for smack-sensing by: Rebecca Fiebrink
// based on code by: Christian Klein
//
// requires: SMS-enabled powerbook or macbook running OS X (10.4.8 or higher)
//
// to compile (in command line):
//     %> make
//
// to run (in command line):
//     %> ./smack-sense [p [t]]
// where first arg, p, is polling period in seconds (e.g., .1 for 10Hz polling)
// and second arg, t, is threshold for acknowledging smack (somewhere between 1 and 25, probably)
//
// If smacking doesn't work, verify that SMS is turned on, by running 
//     "pmset -g" at the command prompt and verifying that "sms" is set to "1".
//  If SMS is not turned on, enable it by running 
//     "sudo pmset -a sms 1" at the prompt.
//-----------------------------------------------------------------------------

/* ********************************************************
 * Original header:
 *
 * a little program to display the coords returned by
 * the powerbook motion sensor
 *
 * A fine piece of c0de, brought to you by
 *
 *               ---===---
 * *** teenage mutant ninja hero coders ***
 *               ---===---
 *
 * All of the software included is copyrighted by Christian Klein <chris@5711.org>.
 *
 * Copyright 2005 Christian Klein. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by Christian Klein.
 * 4. The name of the author must not be used to endorse or promote
 *    products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
 

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <IOKit/IOKitLib.h>

struct data {
    char x;
    char y;
    char z;
    char pad[57];
};

// global handle to SMS data port
io_connect_t  dataPort;

/*
// device names to try
const char * devnames[] = {
    "IOI2CMotionSensor",
    "PMUMotionSensor",
    "SMCMotionSensor"
};

// corresponding data type
const int devtypes[] = {
    kSMSPowerbookDataType,
    kSMSPowerbookDataType,
    kSMSMacBookProDataType
};

// number of things
const int numdevs = sizeof(devnames) / sizeof(const char *);
*/

// signal handler to catch ctrl-c
void signal_int( int sig_num )
{
    fprintf( stderr, "closing...\n" );
    // gracefully exit, maybe
    IOServiceClose(dataPort);
    // bye
    exit(2);
}

// everything in here
int main( int argc, char **argv )
{
    // default duration is .1 second
    int dur = (argc >= 2) ? atoi(argv[1]) : .1;
    // default threshold is 10 (sensitivity)
    int threshold = (argc >= 2) ? atoi(argv[2]) : 10;
    // minimum pause (in seconds) between smack detections
    int recover_time = .2;
    
    // SMS API stuff
    kern_return_t result;
    mach_port_t masterPort;
    io_iterator_t iterator;
    io_object_t aDevice;

    IOItemCount structureInputSize;
    IOByteCount structureOutputSize;

    struct data inputStructure;
    struct data outputStructure;

    // more appley fun
    result = IOMasterPort(MACH_PORT_NULL, &masterPort);
    CFMutableDictionaryRef matchingDictionary = IOServiceMatching("IOI2CMotionSensor");
    result = IOServiceGetMatchingServices(masterPort, matchingDictionary, &iterator);
    if (result != KERN_SUCCESS)
    {
        fputs("IOServiceGetMatchingServices returned error.\n", stderr);
        return 0;
    }

    // what is this doing? (just copy this section)
    aDevice = IOIteratorNext(iterator);
    IOObjectRelease(iterator);
    if(aDevice == 0)
    {
        fputs("No motion sensor available\n", stderr);
        return 0;
    }

    // signal
    signal( SIGINT, signal_int );

    // keep going...
    result = IOServiceOpen(aDevice, mach_task_self(), 0, &dataPort);
    IOObjectRelease(aDevice);
    if(result != KERN_SUCCESS)
    {
        fputs("Could not open motion sensor device\n", stderr);
        return 0;
    }

    // clear out memory
    structureInputSize = sizeof(struct data);
    structureOutputSize = sizeof(struct data);
    memset(&inputStructure, 0, sizeof(inputStructure));
    memset(&outputStructure, 0, sizeof(outputStructure));

    // grabbing the data and making sure it's working...
    result = IOConnectMethodStructureIStructureO(
            dataPort,
            21, // if this doesn't work, try 5 or 24 (SMS API still in flux)
            structureInputSize,
            &structureOutputSize,
            &inputStructure,
            &outputStructure
    );

    // didn't work
    if(result != KERN_SUCCESS)
    {
        puts("SMS feature not enabled or not working");
    }
    else
    {
        // prepare to poll
        int init_x, last_x, init_y, last_y, init_z, last_z;
        init_x = last_x = outputStructure.x;
        init_y = last_y = outputStructure.y;
        init_z = last_z = outputStructure.z;       

        // poll loop
        while (true)
        {
            // poll!
            result = IOConnectMethodStructureIStructureO(
                dataPort,
                21, // this should match the number in above call to IOConnectMethodStructureIStructureO(...)
                structureInputSize,
                &structureOutputSize,
                &inputStructure,
                &outputStructure
            );

            // look at relative displacement compared to last time      
            int diff = (abs(outputStructure.x - last_x) + 
                        abs(outputStructure.y - last_y) + 
                        abs(outputStructure.z - last_z));

            if (diff > threshold)
            {
                // smack detected!  (do stuff here)
                fprintf( stderr, "BANG!\n" );
                
                // remember for next time
                last_x = init_x;
                last_y = init_y;
                last_z = init_z;
                
                // impose minimum wait
                sleep(recover_time);
                
            }
            else
            {
                // no smack detected, remember values for next time
                last_x = outputStructure.x;
                last_y = outputStructure.y;
                last_z = outputStructure.z;
                
                // impose poll rate
                sleep(dur);
            }           
        }   
    }

    // close the device
    IOServiceClose(dataPort);

    return 0;
}
